c!    Program to build domino clusters subject to the rule that we want to
c!    maximise contacting perimeter, and subject to thermal randomisation.

c!    This is essentially Fortran 77 code but has only every been compiled using
c!    gfortran. Here is the output of gfortran --version that worked (Red Hat
c!    Linux):

c!    $ gfortran --version
c!    GNU Fortran (GCC) 4.4.7 20120313 (Red Hat 4.4.7-17)

c!    The script compile.sh should be included with this code, along with the
c!    toolbox programs drawps.f ps_routines.f rannum.f

c!    D.J.Goossens, darren.goossens@gmail.com

c!    November 2016

c!    Version 1.0
 
      program dom2016

      implicit none

       integer ishift(4,2),xsize, ysize,nloop,xcentre, ycentre,deficit
       integer i,j,array2(256,256),cells(256*256,4),shift,numinit
       integer ish,ii,jj,templist(256*256,4),iij,inew,iloop,issh,iiij
       integer dcount,ncell,iq,kq,coordempty(2,256*256),perim(256*256)
       real r(1),temper
       character*1 halt,init(50),verb,pics
       integer comperim,cdmperim,iii,jjj,perimax,sep,sepopt(2)
       integer isdup, idup

c! Relative coords of neighbours
      ishift(1,1) = 1
      ishift(1,2) = 0
      ishift(2,1) = -1
      ishift(2,2) = 0
      ishift(3,1) = 0
      ishift(3,2) = 1
      ishift(4,1) = 0
      ishift(4,2) = -1

c! Ask some questions...

      write(6,*)'xsize and ysize?'
      read(5,*)xsize,ysize
      write(6,*)'size is ',xsize,ysize
      write(6,*)'loops (<0 for autocalc)?'
      read(5,*)nloop
      if(nloop.lt.0) then
         nloop = xsize * ysize
         write(6,*)'Loops set negative; will use ',nloop
      end if
      write(6,*)'Number loops: ',nloop
      halt='y'
      write(6,*)'temperature?'
      read(5,*)temper
      write(6,*)'temperature?',temper
    
      numinit = 2
2023  write(6,*)'Number of initial dominoes? (1 or 2)'
      read(5,*)numinit
      if((numinit.ne.1).and.(numinit.ne.2))goto 2023
      write(6,*)'Number of initial dominoes?',numinit
      do i = 1,numinit
        write(6,*)'Initial domino ',i,'  h or v?'
        read(5,*)init(i)
        write(6,*)'Initial domino ',i,'  h or v?',init(i)
      end do
      sep=0
      if(numinit.eq.2) then
        write(6,*)'Separation?'
        read(5,*)sep
        write(6,*)'Separation?',sep
        write(6,*)'Sep in x, y or x+y? -- (1 0) (0 1) or (1 1)'
        read(5,*)sepopt(1),sepopt(2)
        write(6,*)'Sep in x, y or x+y? -- (1 0) (0 1) or (1 1)',
     &    sepopt(1), sepopt(2)
      end if
      write(6,*)'Verbose output? y/n'
      verb = 'n'
      read(5,*)verb
      write(6,*)'Verbose output? y/n ',verb
      write(6,*)'Pic every cycle y/n'
      pics = 'n'
      read(5,*)pics
      write(6,*)'Pic every cycle y/n ',pics
      write(6,*)'--------------------------------------------------'
      write(6,*)'----------Finished asking for input...------------'
      write(6,*)'--------------------------------------------------'
      write(6,*)nloop,' cycles on a ',xsize,' x',ysize
      write(6,*)'simn read out to fort.2 and fort.3'
      if(halt.eq.'y') then
         write(6,*)'Will halt on hitting edge'
      else
         write(6,*)'Will NOT halt on hitting edge'
      end if
      xcentre = xsize/2
      ycentre = ysize/2
      sep = sep / 2
      do i = 1,256*256
         do j = 1,4
            cells(i,j) = 0
         end do
      end do

c!     array2() keeps track of the individual dominos.  It numbers the
c!     squares with the domino number, with a prefix of + for a horizontal
c!     one and - for a vertical.

      do i = 1,xsize
         do j = 1,ysize
            array2(i,j) = 0
         end do
      end do
      call rseed(11345,21221)

      dcount = 0

c!     Choose starting domino

      dcount = 0
      do i = 1,numinit
      write(6,*)'Setting starting domino ',i
      dcount = dcount +1
         ii = xcentre + sepopt(1) * sep * (2*i-3)
         jj = ycentre + sepopt(2) * sep * (2*i-3)
         if(init(i).eq.'h')then
            array2(ii,jj) =  1
         else
            array2(ii,jj) = -1
         end if
!      end if
      if(array2(ii,jj).eq.1)
     &     write(6,*)'Initial domino horizontal',i
      if(array2(ii,jj).eq.-1)
     &     write(6,*)'Initial domino vertical',i

c!    cells() tells us which cells belong ot each domino

c!    Choose which is next cell to fill

      call rannum(r,1)
      shift = 2 * nint(r(1)) - 1
      if (array2(ii,jj).eq.1) then
         array2(ii+shift,jj) = array2(ii,jj)
         write(6,*)'Initial at ',ii,jj,ii+shift,jj
         cells(dcount,1) = ii
         cells(dcount,2) = jj
         cells(dcount,3) = ii+shift
         cells(dcount,4) = jj
      else
         array2(ii,jj+shift) = array2(ii,jj)
         write(6,*)'Initial at ',ii,jj,ii,jj+shift
         cells(dcount,1) = ii
         cells(dcount,2) = jj
         cells(dcount,3) = ii
         cells(dcount,4) = jj+shift
      end if
      end do

      if(pics.eq.'y') then
         call drawerfileps(array2,xsize,ysize,dcount)
         call drawerfile(array2,xsize,ysize,dcount)
      end if
      write(6,*)'--------------------------------------------------'
      write(6,*)'-----Finished initialising.  Entering loop...-----'
      write(6,*)'--------------------------------------------------'
      write(6,*)''

c!    OK, step 1 is 'Find all possible new dominoes' and pay attention to
c!    adjoining perimeter.

       do iloop=1,nloop
       write(6,'(a)',advance='no')'-'
   
       if(verb.eq.'y')write(6,*)iloop,' is cycle -------------------'
c!     List of cells with filled neighbours (i.e., perimeter cells)
       ncell = 0
         do i=1+1,xsize-1
           do j=1+1,ysize-1
             if(array2(i,j).eq.0) then
c!             Does it have a filled neighbour?
               if((array2(i+1,j).ne.0).or.(array2(i-1,j).ne.0)
     &         .or.(array2(i,j+1).ne.0).or.
     &         (array2(i,j-1).ne.0))then
                 if(verb.eq.'y')write(6,*)i,j,' has a neighbour'
c!               if we are at an edge and we have a neighbour, stop!
c!               This just means, if we are inside the if() and i or j is 1 or x/ysize, then stop.
                 if(((j+1).eq.ysize).or.((i+1).eq.xsize).or.
     &           ((j-1).eq.1).or.((i-1).eq.1)) then
c!                 do the stuff we do at the end and stop
                   write(6,'(a6)')'> Done'
                   write(6,*)array2(i+1,j),array2(i-1,j),
     &                       array2(i,j+1),array2(i,j-1)
                   write(6,*)'Outputs, ps files, pictures and data'
                   call drawerfileps(array2,xsize,ysize,dcount)
                   call drawerfile(array2,xsize,ysize,dcount)
                   do iq = 1,xsize
                     do  kq=1,ysize
                        write(2,'(4i8)')iq,kq,array2(iq,kq)
                     end do
                   end do
                   do iq = 1, dcount
                     write(3,'(7i5)')iq,(cells(iq,kq),kq=1,4)
                   end do
                   stop
                 end if
                 ncell = ncell + 1
                 coordempty(1,ncell) = i
                 coordempty(2,ncell) = j
               end if
             end if
           end do
         end do
         if(verb.eq.'y')then
           write(6,*)ncell,' empty cells have filled neighbours.'
           write(6,*)'They are:'
           do i=1,ncell
             write(6,*)coordempty(1,i),coordempty(2,i)
           end do
         end if

c!       OK, so we have a list of cells with neighbours.  Now we need to compile a list
c!       of possible new dominos, and get how much perimeter each shares with exisiting ones.

c!       Go to a cell on the perimeter. Go to a cell beside it. See if that cell is empty.
c!       if so, work out the perimeter this domino has in common with the cluster.

c!       This will be a 2 step process. Need to work out the perimeters, then work out
c!       the maximum then choose randomly one to add, subject to temperature.

         perimax=-3
         inew = 0
         do iij = 1, ncell
           if(verb.eq.'y')write(6,*)'---------------------------------'
           i = coordempty(1,iij)
           j = coordempty(2,iij)
c!         How many of the squares around i,j are occupied?
           comperim=0
           do ish=1,4
             ii = i + ishift(ish,1)
             jj = j + ishift(ish,2)
             if(array2(ii,jj).ne.0)comperim=comperim+1
           end do
           if(verb.eq.'y')write(6,*)'square, neighbours: ',i,j,comperim
c!         Go to each empty cell, look at its four neighbours, see which are empty,
           do ish = 1,4
             ii = i + ishift(ish,1)
             jj = j + ishift(ish,2)
c !          if cell is empty:
             if(array2(ii,jj).eq.0) then
c!             if it is in the perimeter it is a possible domino
c!             so work out the common perimeter of the new domino
c!             this is the same as the number of occupied surrounding squares.                  
c!             but first, we check if it is a duplicate
               isdup = 0
               do idup = 1,inew
                 if((i.eq.templist(idup,1)).and.
     &            (j.eq.templist(idup,2)).and.
     &            (ii.eq.templist(idup,3)).and.
     &            (jj.eq.templist(idup,4)))isdup=1
                 if((i.eq.templist(idup,3)).and.
     &            (j.eq.templist(idup,4)).and.
     &            (ii.eq.templist(idup,1)).and.
     &            (jj.eq.templist(idup,2)))isdup=1
               end do
               if(isdup.eq.0) then

                 cdmperim=0
                 do issh=1,4
                   iii = ii + ishift(issh,1)
                   jjj = jj + ishift(issh,2)
                   if(array2(iii,jjj).ne.0)cdmperim=cdmperim+1
                 end do
                 if(verb.eq.'y')write(6,*)
     &                   'square, neighbours: ',ii,jj,cdmperim   

c!               Domino!
                 inew = inew + 1
                 perim(inew)=cdmperim+comperim
                 if(perim(inew).gt.perimax)perimax=perim(inew)
                 if(verb.eq.'y')then
                   write(6,*)'overall option number', inew
                   write(6,*) i,j,ii,jj,perim(inew)
                 end if
c!               add to templist
                 templist(inew,1)=i
                 templist(inew,2)=j
                 templist(inew,3)=ii
                 templist(inew,4)=jj
               end if
             end if
           end do
         end do
         if(verb.eq.'y')then
           write(6,*)inew,' possible dominoes at:'
           do i = 1,inew
             write(6,*)(templist(i,j),j=1,4),perim(i)
           end do
           write(6,*)'maximum perimeter is',perimax
         end if

101      call rannum(r,1)
         iiij = int(inew * r(1)) +1
         deficit = perimax - perim(iiij)
         call rannum(r,1)
         if(exp(-1.0*real(deficit)/temper).lt.r(1))goto 101
c!       So the line above just sends the program back to choose another option
c!       if the random number is bigger than the exponential. It could be the same
c!       domino, but with a different exponential it might get through...!

         if(verb.eq.'y')write(6,*)'It is number ',iiij
         dcount=dcount+1
c!       Is is a horizontal or vertical domino? If i = ii, it is vertical... 
         if(templist(iiij,1).eq.templist(iiij,3))then
             array2(templist(iiij,1),templist(iiij,2))=-1*dcount
             array2(templist(iiij,3),templist(iiij,4))=-1*dcount
         else
             array2(templist(iiij,1),templist(iiij,2))=dcount
             array2(templist(iiij,3),templist(iiij,4))=dcount
         end if
         do i = 1,4
             cells(dcount,i)=templist(iiij,i)
         end do
c!    Here is where the ascii and PostScript art comes in...
      if(pics.eq.'y') then
         call drawerfileps(array2,xsize,ysize,dcount)
         call drawerfile(array2,xsize,ysize,dcount)
      end if
c!    loop!
      end do 
      write(6,'(a6)')'> Done'
                        write(6,*)array2(i+1,j),array2(i-1,j),
     &                       array2(i,j+1),array2(i,j-1)
                        write(6,*)'Outputs, ps files, pictures and data'
                        call drawerfileps(array2,xsize,ysize,dcount)
                        call drawerfile(array2,xsize,ysize,dcount)
                        do iq = 1,xsize
                           do  kq=1,ysize
                              write(2,'(4i8)')iq,kq,array2(iq,kq)
                           end do
                        end do
                        do iq = 1, dcount
                           write(3,'(7i5)')iq,(cells(iq,kq),kq=1,4)
                        end do

      stop

      end program dom2016

c!     -----------------------------------------------------------------

      subroutine drawerfile(array,xsize,ysize,nfile)

      implicit none

      integer array(256,256)
      character*1 oneline(258)
      integer xsize,ysize
      integer i,j,nfile,unum
      character*20 fname

      unum = nfile
      if((unum.lt.10))write(fname,'(a8,i1)')'pic.0000',unum
      if((unum.gt.9).and.(unum.lt.100))
     &     write(fname,'(a7,i2)')'pic.000',unum 
      if((unum.lt.1000).and.(unum.gt.99))
     &     write(fname,'(a6,i3)')'pic.00',unum 
      if((unum.lt.10000).and.(unum.gt.999))
     &     write(fname,'(a5,i4)')'pic.0',unum 
      if((unum.lt.100000).and.(unum.gt.9999))
     &     write(fname,'(a4,i5)')'pic.',unum 
      do i = 1,xsize + 2
         oneline(i)='_'
      end do
      open(file=fname,unit = 1)
      write(1,*)(oneline(i),i=1,xsize+2)
      do j = 1, ysize
         do i = 1,xsize
            if(array(i,j).le.-1) then
               oneline(i)='O'
            else
               if(array(i,j).ge.1) then
                  oneline(i)='o'
               else
                  oneline(i)=' '
               end if
            end if
         end do
         write(1,*)'|',(oneline(i),i=1,xsize),'|'
      end do
      do i = 1,xsize + 2
         oneline(i)='~'
      end do
      write(1,*)(oneline(i),i=1,xsize+2)

      close(1)

      return

      end subroutine drawerfile

c!     -----------------------------------------------------------------

